"""
Then main unit tests.
"""
from io import StringIO
import os
from django.conf import settings
from django.test import TestCase
from django.core.management import call_command

from django_cenvars import models
from django_cenvars.tools import codec

settings.RSA_KEYSIZE = 512 # For testing purposes, make it a bit faster.

# pylint: disable=no-member
class TestMain(TestCase):
    "Main testing."
    def _create_environments(self):
        "Create some data"
        gen1 = models.Environment.objects.create(label='gen1')
        gen2 = models.Environment.objects.create(label='gen2')
        gen3 = models.Environment.objects.create(label='gen3')
        gen4 = models.Environment.objects.create(label='gen4')
        self.assertIn('| gen1', str(gen1))
        return gen1, gen2, gen3, gen4

    def _create_inheritance(self, gen1, gen2, gen3, gen4):
        "Create inheritance."
        g1_2 = models.Inheritance.objects.create(offspring=gen2, ascendant=gen1)
        g2_3 = models.Inheritance.objects.create(offspring=gen3, ascendant=gen2)
        g3_4 = models.Inheritance.objects.create(offspring=gen4, ascendant=gen3)
        self.assertIn('gen2<gen1', str(g1_2))
        return g1_2, g2_3, g3_4

    def _create_variabels(self, gen1, gen2, gen3, gen4):
        "Create the variables."
        vg1 = gen1.add_variable('gen1k', 'gen1_value')
        vg2 = gen2.add_variable('gen2k', 'gen2_value')
        vg3 = gen3.add_variable('gen3k', 'gen3_value')
        vg4 = gen4.add_variable('gen4k', 'gen4_value')
        self.assertIn('gen1k', str(vg1[0]))
        self.assertIn('gen1 | gen1k | gen1_value', str(vg1[1]))
        return vg1, vg2, vg3, vg4


    def test_001_smoke(self):
        "models.environment"
        environ = models.Environment.objects.create(label='test')
        self.assertTrue(environ.envar is not None)
        data = os.environ.copy()
        _, rsa_key = codec.decode_key(environ.envar)
        encrypted = codec.encrypt(rsa_key, data)
        decrypted = codec.decrypt(rsa_key, encrypted)
        self.assertEqual(data, decrypted)

    def test_002_ancestors(self):
        "Test the inherited"
        env_a = models.Environment.objects.create(label='a')
        env_b = models.Environment.objects.create(label='b')
        env_c = models.Environment.objects.create(label='c')
        models.Inheritance.objects.create(offspring=env_c, ascendant=env_b)
        models.Inheritance.objects.create(offspring=env_b, ascendant=env_a)
        # C is a child of B which is a child of A, so when we resolve the
        # ancestors of C we expect it to return C, B, A
        self.assertEqual([env.label for env in (env_c, env_b, env_a)],
                         [env.label for env in env_c.resolve_ancestors()])

    def test_003_inheritance(self):
        "Test the inheritance"
        gen1, gen2, gen3, gen4 = self._create_environments()
        self._create_inheritance(gen1, gen2, gen3, gen4)
 
        expected = [gen4, gen3, gen2, gen1]
        expected = [str(item) for item in expected]
        returned = gen4.resolve_ancestors()
        returned = [str(item) for item in returned]
        self.assertEqual(expected, returned)
 
        # Add a recursion in it
        models.Inheritance.objects.create(offspring=gen1, ascendant=gen4)
        returned = gen4.resolve_ancestors()
        returned = [str(item) for item in returned]
        self.assertEqual(expected, returned)
 
    def test_004_variables(self):
        "Test retrieving variables"
        gen1, gen2, gen3, gen4  = self._create_environments()
        self._create_inheritance(gen1, gen2, gen3, gen4)
        self._create_variabels(gen1, gen2, gen3, gen4)
        gen1.add_variable('gen1_override', 'first')
        gen4.add_variable('gen1_override', 'override')
 
        self.assertEqual(gen4.get_variables()['gen1_override'], 'override')
        self.assertEqual(gen3.get_variables()['gen1_override'], 'first')

    def test_005_fetch_via_request(self):
        "Implement the fetching of the data via the get request."
        gen1, gen2, gen3, gen4  = self._create_environments()
        self._create_inheritance(gen1, gen2, gen3, gen4)
        self._create_variabels(gen1, gen2, gen3, gen4)
        encrypted = self.client.get('/cenvars/%s/' % gen4.ident).content
        rsa_key = codec.decode_key(gen4.envar)[1]
        data = codec.decrypt(rsa_key, encrypted)
        self.assertEqual(gen4.get_variables(), data)

    def test_006_models_save(self):
        "test if we can set the label without regenerating a key."
        gen1 = self._create_environments()[0]
        old = gen1.ident[::]
        gen1.label = 'Generation 1'
        gen1.save()
        self.assertEqual(gen1.ident, old)

    def test_007_command(self):
        "test if the command has output."
        stdout = StringIO()
        call_command('cenvars_newkey', stdout=stdout)
        self.assertTrue(len(stdout.getvalue()) > 0)

    def test_008_codec(self):
        "test the codecs"
        rsa_key = codec.decode_key(settings.CENVARS_KEY)[1]
        data = 'test'
        cypher = codec.encrypt(rsa_key, data)
        plain = codec.decrypt(rsa_key, cypher)
        self.assertEqual(data, plain)

    def test_009_save_load(self):
        "test if the data is persistent saved and restored"
        self._create_environments()
        from ..models import Environment
        environments = Environment.objects.all()
        self.assertTrue(environments.exists())

        from django_memdb.models import PersistentStorage
        query = list(PersistentStorage.objects.all())
        environments.delete()
        self.assertFalse(Environment.objects.all().exists())

        settings.MEMDB_RESTORED = False
        from django_memdb.tools.restore import restore
        restore(query)

        self.assertTrue(Environment.objects.all().exists())



"""
This module is imported on app ready (see __init__).
"""
from django.conf import settings
from django_memdb.signals import store_load, store_save
from . import __info__
from .tools import codec

if settings.CENVARS_KEY is None: # pragma: no cover
    KEY = None
else:
    KEY = codec.decode_key(settings.CENVARS_KEY)[1]

def persistent_crypt(**kwargs):
    "Persistent data for memdb is stored encrypted."
    # We only want to do crypto for the models in this application.
    if __info__.LABELS['name'] == kwargs['kwargs']['application']:
        data = kwargs['kwargs']['data']
        if kwargs['kwargs']['process'] == 'encode':
            data = codec.encrypt(KEY, data)
        else:
            data = codec.decrypt(KEY, data)

        kwargs['kwargs']['data'] = data

store_load.connect(persistent_crypt)
store_save.connect(persistent_crypt)
 


# -*- coding: utf-8 -*-
# pylint: disable=invalid-name, missing-docstring, line-too-long
# Generated by Django 1.10.1 on 2016-09-14 20:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_memdb.mixins


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Environment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(blank=True, default='', max_length=64)),
                ('store', models.URLField(blank=True)),
                ('ident', models.CharField(blank=True, max_length=56, unique=True)),
                ('envar', models.TextField(blank=True)),
            ],
            bases=(models.Model, django_memdb.mixins.PeristentInMemoryDB),
        ),
        migrations.CreateModel(
            name='Inheritance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ascendant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ascendant', to='django_cenvars.Environment')),
                ('offspring', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='offspring', to='django_cenvars.Environment')),
            ],
            bases=(models.Model, django_memdb.mixins.PeristentInMemoryDB),
        ),
        migrations.CreateModel(
            name='Key',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=128, unique=True)),
            ],
            bases=(models.Model, django_memdb.mixins.PeristentInMemoryDB),
        ),
        migrations.CreateModel(
            name='Variable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.TextField(blank=True, null=True)),
                ('environment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_cenvars.Environment')),
                ('key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_cenvars.Key')),
            ],
            bases=(models.Model, django_memdb.mixins.PeristentInMemoryDB),
        ),
    ]



"""
Envar models.
"""
import json
import random
import hashlib
import base64
import rsa
from django.db import models
from django.conf import settings
from django_memdb.mixins import PeristentInMemoryDB
from ..tools import codec

def is_empty(value):
    "Test if value is empty."
    if value is None or len(value) == 0:
        return True
    return False

# pylint: disable=too-many-instance-attributes, no-member
class Environment(models.Model, PeristentInMemoryDB):
    "A single environment instance."
    label = models.CharField(max_length=64, default='', blank=True)
    store = models.URLField(blank=True)
    ident = models.CharField(max_length=56, unique=True, blank=True)
    envar = models.TextField(blank=True)

    def __str__(self):
        return self.ident + ' | ' + self.label

    def save(self, *args, **kwargs):
        "override save."
        if is_empty(self.store):
            self.store = settings.DEFAULT_SERVER

        if is_empty(self.envar):
            _, key = rsa.newkeys(settings.RSA_KEYSIZE)
            self.ident, self.envar = codec.encode_key(key, self.store)

        return models.Model.save(self, *args, **kwargs)

    def resolve_ancestors(self):
        "Resolve inheritance, break if there is a recursion."
        # Get a chronological line of inheritance, preventing infinite loops.
        inheritances = []
        work_queue = [self,]
        while len(work_queue) > 0:
            environment = work_queue.pop(0)
            if environment not in inheritances:
                inheritances.append(environment)
            for inheritance in environment.offspring.all():
                if inheritance.ascendant not in inheritances+work_queue:
                    work_queue.append(inheritance.ascendant)

        return inheritances

    def get_variables(self):
        "Return all variables"
        tmp = dict()
        inheritances = self.resolve_ancestors()
        inheritances.reverse()
        for environment in inheritances:
            for variable in environment.variable_set.all():
                tmp[variable.key.key] = variable.value

        return tmp

    def add_variable(self, key, value):
        "Add a variable associated to this environment"
        key = Key.objects.get_or_create(key=key)[0]
        var = Variable.objects.create(environment=self, key=key, value=value)
        return key, var



class Key(models.Model, PeristentInMemoryDB):
    "Key, well it is a key."
    key = models.CharField(max_length=128, unique=True)

    def __str__(self):
        return self.key


class Variable(models.Model, PeristentInMemoryDB):
    "The variable."
    environment = models.ForeignKey(Environment)
    key = models.ForeignKey(Key)
    value = models.TextField(blank=True, null=True)

    def __str__(self):
        _ = self.environment.label + ' | ' + self.key.key + ' | ' + self.value
        return  _


class Inheritance(models.Model, PeristentInMemoryDB):
    "Inheritance of Environments."
    offspring = models.ForeignKey(Environment, related_name='offspring')
    ascendant = models.ForeignKey(Environment, related_name='ascendant')

    class Meta: # pylint:disable=too-few-public-methods, missing-docstring
        unique_together = (('offspring', 'ascendant'))

    def __str__(self):
        return self.offspring.label + '<' + self.ascendant.label




"""
Command update languages.
"""
import rsa
from django.conf import settings
from django.core.management.base import BaseCommand
from ...tools import codec

def create():
    "create key"
    _,key = rsa.newkeys(settings.RSA_KEYSIZE)
    # need to split this out in a encode/decode key
    _, tmp = codec.encode_key(key, url=settings.DEFAULT_SERVER)
    return tmp


class Command(BaseCommand):
    """Create a new server key pair."""
    help = "Create a new server key pair."

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        return create()




"""
Codecs used for keys, etc.
"""
import json
import zlib
import base64
import rsa
import pyaes
import os
import hashlib
from django.conf import settings


def encrypt(rsa_key, data):
    "Encrypt/Encode data dictionary using json, utf8, zlib, aes and rsa."
    data_json = json.dumps(data)
    data_utf8 = data_json.encode('utf8')
    data_zlib = zlib.compress(data_utf8, 9)
    key_aes = os.urandom(32)
    aes = pyaes.AESModeOfOperationCTR(key_aes)
    data_aes = aes.encrypt(data_zlib)
    rsa_aes = rsa.encrypt(key_aes, rsa_key)
    returns = rsa_aes+data_aes
    return returns

def decrypt(rsa_key, encrypted):
    "Decrypt/Decode to data dictionary using rsa, aes, zlib, utf8 and json."
    size = settings.RSA_KEYSIZE // 8
    rsa_aes = encrypted[:size]
    encrypted = encrypted[size:]
    key_aes = rsa.decrypt(rsa_aes, rsa_key)
    aes = pyaes.AESModeOfOperationCTR(key_aes)
    data_aes = aes.decrypt(encrypted)
    data_zlib = zlib.decompress(data_aes)
    data_utf8 = data_zlib.decode('utf8')
    data = json.loads(data_utf8)
    return data

def encode_key(decrypt_key, url):
    "Encode key and url to hexdigest and environment variable."
    tmp = dict()
    tmp['u'] = url
    tmp['n'] = decrypt_key.n
    tmp['e'] = decrypt_key.e
    tmp['d'] = decrypt_key.d
    tmp['p'] = decrypt_key.p
    tmp['q'] = decrypt_key.q

    tmp_json = json.dumps(tmp)
    tmp_utf8 = tmp_json.encode('utf8')
    tmp_zlib = zlib.compress(tmp_utf8, 9)
    tmp_b64e = base64.b64encode(tmp_zlib)
    tmp_envk = tmp_b64e.decode('utf8')
    n_bytes = decrypt_key.n.to_bytes(settings.RSA_KEYSIZE // 8, 'big')
    digest = hashlib.sha224(n_bytes).hexdigest().upper()
    return digest, tmp_envk

def decode_key(encoded):
    "Decode environment variable to url and key."
    tmp_byte = encoded.encode('utf8')
    tmp_b64e = base64.b64decode(tmp_byte)
    tmp_zlib = zlib.decompress(tmp_b64e)
    tmp_utf8 = tmp_zlib.decode('utf8')
    tmp = json.loads(tmp_utf8)
    key = rsa.PrivateKey(tmp['n'], tmp['e'], tmp['d'], tmp['p'], tmp['q'])
    return tmp['u'], key



"""
Project Views
"""

from django.shortcuts import get_object_or_404
from django.http import HttpResponse
from ..models import Environment
from ..tools.codec import decode_key, encrypt

# pylint: disable=unused-argument
def view(request, identifier):
    "Example view"
    environment = get_object_or_404(Environment, ident=identifier)
    data = environment.get_variables()
    rsa_key = decode_key(environment.envar)[1]
    encrypted = encrypt(rsa_key, data)
    return HttpResponse(content=encrypted)




"""
Default admin interface setup.
"""
from django.contrib import admin
from .tools import models

# First define custom admin on required models.

# Now register the remaining with the default settings.
for key in models.ALL:
    subject = models.ALL[key]

    if not admin.site.is_registered(subject):
        admin.site.register(subject)


"""
PyPi Setup file.
"""
# pylint: disable=no-name-in-module, import-error
from setuptools import setup, find_packages

AUTHOR = 'Martin P. Hellwig'
AUTHOR_EMAIL = 'martin.hellwig@gmail.com'

NAME = 'django-cenvars'
DESCRIPTION = 'Django Centralised Environment Variables Service'
VERSION = '0.0.0.6'

URL_MAIN = "https://bitbucket.org/hellwig/" + NAME + '/'
URL_DOWNLOAD = URL_MAIN + 'download/' + VERSION + '.zip'


KEYWORDS = [
    'django',
    'django-integrator'
    ]

CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    ]

REQUIREMENTS = [
    'Django',
    'django-integrator',
    'pyaes',
    'rsa',
    'django-memdb',
    ]

LICENSE = 'BSD'

################################################################################

KWARGS = {
    'name':NAME, 'packages':find_packages(), 'version':VERSION,
    'description':DESCRIPTION, 'author':AUTHOR, 'author_email':AUTHOR_EMAIL,
    'url':URL_MAIN, 'download_url':URL_DOWNLOAD, 'keywords':KEYWORDS,
    'license':LICENSE, 'classifiers':CLASSIFIERS,
    'install_requires':REQUIREMENTS}

setup(**KWARGS)



"""
This projects settings.
"""
# These settings can be imported with the django_app_importer tool
import os

URLCONF = 'urls'
DEFAULT_SERVER = os.environ.get('CENVARS_URL', 'https://localhost:8000')
RSA_KEYSIZE = 2048

CENVARS_KEY = os.environ.get('CENVARS_KEY')

CENVARS_DATABASE = 'cenvars'

DATABASES = {
    CENVARS_DATABASE: {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'file::memory:?cache=shared'}
}

# pylint:disable=wrong-import-position
import django_integrator
django_integrator.add_application('django_memdb')
